@@ -126,12 +126,6 @@ class ArrayStrategy(SearchStrategy):
     def __init__(self, element_strategy, shape, dtype, fill, unique):
         self.shape = tuple(shape)
         self.fill = fill
-        assert shape, "Zero-dimensional array shape is special-cased in arrays()"
-        check_argument(
-            all(isinstance(s, integer_types) for s in shape),
-            "Array shape must be integer in each dimension, provided shape was {}",
-            shape,
-        )
         self.array_size = int(np.prod(shape))
         self.dtype = dtype
         self.element_strategy = element_strategy
@@ -298,11 +292,10 @@ def fill_for(elements, unique, fill, name=""):
     return fill
 
 
-@st.composite
+@st.defines_strategy
 def arrays(
-    draw,  # type: Any
     dtype,  # type: Any
-    shape,  # type: Union[int, Sequence[int]]
+    shape,  # type: Union[int, Sequence[int], st.SearchStrategy[Sequence[int]]]
     elements=None,  # type: st.SearchStrategy[Any]
     fill=None,  # type: st.SearchStrategy[Any]
     unique=False,  # type: bool
@@ -310,12 +303,11 @@ def arrays(
     # type: (...) -> st.SearchStrategy[np.ndarray]
     r"""Returns a strategy for generating :class:`numpy:numpy.ndarray`\ s.
 
-    * ``dtype`` may be any valid input to
-      :class:`numpy.dtype <numpy:numpy.dtype>` (this includes
-      :class:`~numpy:numpy.dtype` objects), or a strategy that generates such
-      values.
-    * ``shape`` may be an integer >= 0, a tuple of length >= 0 of such
-      integers, or a strategy that generates such values.
+    * ``dtype`` may be any valid input to :class:`~numpy:numpy.dtype`
+      (this includes :class:`~numpy:numpy.dtype` objects), or a strategy that
+      generates such values.
+    * ``shape`` may be an integer >= 0, a tuple of such integers, or a
+      strategy that generates such values.
     * ``elements`` is a strategy for generating values to put in the array.
       If it is None a suitable value will be inferred based on the dtype,
       which may give any legal value (including eg ``NaN`` for floats).
@@ -378,28 +370,41 @@ def arrays(
     hundreds or more elements, having a fill value is essential if you want
     your tests to run in reasonable time.
     """
+    # We support passing strategies as arguments for convenience, or at least
+    # for legacy reasons, but don't want to pay the perf cost of a composite
+    # strategy (i.e. repeated argument handling and validation) when it's not
+    # needed.  So we get the best of both worlds by recursing with flatmap,
+    # but only when it's actually needed.
     if isinstance(dtype, SearchStrategy):
-        dtype = draw(dtype)
+        return dtype.flatmap(
+            lambda d: arrays(d, shape, elements=elements, fill=fill, unique=unique)
+        )
+    if isinstance(shape, SearchStrategy):
+        return shape.flatmap(
+            lambda s: arrays(dtype, s, elements=elements, fill=fill, unique=unique)
+        )
+    # From here on, we're only dealing with values and it's relatively simple.
     dtype = np.dtype(dtype)
     if elements is None:
         elements = from_dtype(dtype)
-    if isinstance(shape, SearchStrategy):
-        shape = draw(shape)
-    if isinstance(shape, int):
+    if isinstance(shape, integer_types):
         shape = (shape,)
     shape = tuple(shape)
-    if not shape:
-        return np.full(shape=(), fill_value=draw(elements), dtype=dtype)
+    check_argument(
+        all(isinstance(s, integer_types) for s in shape),
+        "Array shape must be integer in each dimension, provided shape was {}",
+        shape,
+    )
     fill = fill_for(elements=elements, unique=unique, fill=fill)
-    return draw(ArrayStrategy(elements, shape, dtype, fill, unique))
+    return ArrayStrategy(elements, shape, dtype, fill, unique)
 
 
 @st.defines_strategy
 def array_shapes(min_dims=1, max_dims=3, min_side=1, max_side=10):
     # type: (int, int, int, int) -> st.SearchStrategy[Tuple[int, ...]]
     """Return a strategy for array shapes (tuples of int >= 1)."""
-    order_check("dims", 1, min_dims, max_dims)
-    order_check("side", 1, min_side, max_side)
+    order_check("dims", 0, min_dims, max_dims)
+    order_check("side", 0, min_side, max_side)
     return st.lists(
         st.integers(min_side, max_side), min_size=min_dims, max_size=max_dims
     ).map(tuple)
